#!/usr/bin/env python
# coding: utf-8

# This script has kindly been provided by Julia Schessner.

from io import BytesIO
from reportlab.pdfgen import canvas
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Paragraph


from pdfrw import PdfReader, PdfDict
from pdfrw.buildxobj import pagexobj
from pdfrw.toreportlab import makerl
from reportlab.platypus import Flowable
from reportlab.lib.enums import TA_JUSTIFY,TA_LEFT,TA_CENTER,TA_RIGHT

# The following class was copied from https://stackoverflow.com/questions/3448365/pdf-image-in-pdf-document-using-reportlab-python (answer from skidzo, 2017)
class PdfImage(Flowable):
    """
    PdfImage wraps the first page from a PDF file as a Flowable
    which can be included into a ReportLab Platypus document.
    Based on the vectorpdf extension in rst2pdf (http://code.google.com/p/rst2pdf/)

    This can be used from the place where you want to return your matplotlib image
    as a Flowable:

        img = BytesIO()

        fig, ax = plt.subplots(figsize=(canvaswidth,canvaswidth))

        ax.plot([1,2,3],[6,5,4],antialiased=True,linewidth=2,color='red',label='a curve')

        fig.savefig(img,format='PDF')

        return(PdfImage(img))

    """

    def __init__(self, filename_or_object, width=None, height=None, kind='direct'):
        # If using StringIO buffer, set pointer to begining
        if hasattr(filename_or_object, 'read'):
            filename_or_object.seek(0)
            #print("read")
        self.page = PdfReader(filename_or_object, decompress=False).pages[0]
        self.xobj = pagexobj(self.page)

        self.imageWidth = width
        self.imageHeight = height
        x1, y1, x2, y2 = self.xobj.BBox

        self._w, self._h = x2 - x1, y2 - y1
        if not self.imageWidth:
            self.imageWidth = self._w
        if not self.imageHeight:
            self.imageHeight = self._h
        self.__ratio = float(self.imageWidth)/self.imageHeight
        if kind in ['direct','absolute'] or width==None or height==None:
            self.drawWidth = width or self.imageWidth
            self.drawHeight = height or self.imageHeight
        elif kind in ['bound','proportional']:
            factor = min(float(width)/self._w,float(height)/self._h)
            self.drawWidth = self._w*factor
            self.drawHeight = self._h*factor

    def wrap(self, availableWidth, availableHeight):
        """
        returns draw- width and height

        convenience function to adapt your image
        to the available Space that is available
        """
        return self.drawWidth, self.drawHeight

    def drawOn(self, canv, x, y, _sW=0):
        """
        translates Bounding Box and scales the given canvas
        """
        if _sW > 0 and hasattr(self, 'hAlign'):
            a = self.hAlign
            if a in ('CENTER', 'CENTRE', TA_CENTER):
                x += 0.5*_sW
            elif a in ('RIGHT', TA_RIGHT):
                x += _sW
            elif a not in ('LEFT', TA_LEFT):
                raise ValueError("Bad hAlign value " + str(a))

        #xobj_name = makerl(canv._doc, self.xobj)
        xobj_name = makerl(canv, self.xobj)

        xscale = self.drawWidth/self._w
        yscale = self.drawHeight/self._h

        x -= self.xobj.BBox[0] * xscale
        y -= self.xobj.BBox[1] * yscale

        canv.saveState()
        canv.translate(x, y)
        canv.scale(xscale, yscale)
        canv.doForm(xobj_name)
        canv.restoreState()


def draw_paragraph(text, pdf, cw, ch, poi, centered_vertically=False):
    P = Paragraph(text)
    w,h = P.wrap(cw, ch)
    if not centered_vertically:
        poi[1] -= h
    else:
        poi[1] -= h+int((ch-h)/2)
    P.drawOn(pdf, poi[0], poi[1])
    poi[0] += w
    return poi

def draw_plotly(fig, pdf, cw, ch, poi, rescale=False, centerv=True, centerh=True,
                rasterize = False, png_scaling=4):
    w = fig.layout.width
    if w is None:
        w = cw
        if rescale:
            fig.update_layout(width=w)
    h = fig.layout.height
    if h is None:
        h = ch
        if rescale:
            fig.update_layout(height=h)
    if centerh:
        poi[0] += int((cw-w)/2)
    if centerv:
        poi[1] -= h+int((ch-h)/2)
    else:
        poi[1] -= h
    if rasterize:
        img = ImageReader(BytesIO(fig.to_image(format='png', scale=png_scaling)))
        pdf.drawImage(img, poi[0], poi[1], width=w, height=h)
    else:
        img = PdfImage(BytesIO(fig.to_image(format='pdf')), width=w, height=h)
        img.drawOn(pdf, poi[0], poi[1])
    poi[0] += w
    return poi


def draw_bytes(b, pdf, cw, ch, poi):
    poi[1] -= ch
    pdf.drawImage(ImageReader(BytesIO(b)), poi[0], poi[1])
    poi[0] += cw
    return poi


def draw_content(pdf, content, width=595, height=842, border=40, spacing=7, png_scaling=4, verbose=False):
    content_width = width-(2*border)
    content_height = height-(2*border)
    pointer = [border, height-border]
    fontsize = pdf._fontsize
    if type(content) == str:
        draw_paragraph(content, pdf, content_width, content_height, pointer, centered_vertically=True)
    elif str(type(content)) == "<class 'plotly.graph_objs._figure.Figure'>":
        draw_plotly(content, pdf, content_width, content_height, pointer, png_scaling=png_scaling)
    elif type(content) == bytes:
        draw_bytes(content, pdf, content_width, content_height, pointer)
    elif type(content) == list:

        # initialize left content-height
        ch = content_height

        for ri, row in enumerate(content):

            if verbose:
                print('row', ri, 'left space:', content_width, ch, 'pointer:', pointer)

            # draw row
            if type(row) == str:
                pointer = draw_paragraph(row, pdf, content_width, ch, pointer)
            elif str(type(row)) == "<class 'plotly.graph_objs._figure.Figure'>":
                pointer = draw_plotly(row, pdf, content_width, ch, pointer,
                                      centerv=False, png_scaling=png_scaling)
            elif type(row) == bytes:
                pointer = draw_bytes(row, pdf, content_width, ch, pointer)
            elif type(row) == list:

                # initialize left content-width and store pointer height, initialize max height
                cw = content_width
                poih = pointer[1]
                poihmax = 0

                for ii, i in enumerate(row):

                    if verbose:
                        print('item', ii, 'left space:', cw, ch, 'pointer:', pointer)

                    # draw item
                    if type(i) == str:
                        pointer = draw_paragraph(i, pdf, cw, ch, pointer)
                    elif str(type(i)) == "<class 'plotly.graph_objs._figure.Figure'>":
                        pointer = draw_plotly(i, pdf, cw, ch, pointer,
                                              centerv=False, centerh=False, png_scaling=png_scaling)
                    elif type(i) == bytes:
                        pointer = draw_bytes(i, pdf, cw, ch if poihmax == 0 else poihmax, pointer)
                    else:
                        pointer = draw_paragraph("Unknown content of {} passed.".format(str(type(i))), pdf,
                                       cw, ch, pointer)

                    # check max height, reset height pointer for next item, raise overflow warning,
                    # add spacing and recalculate leftover content width
                    poihmax = max(poihmax, poih-pointer[1])
                    pointer[1] = poih
                    if pointer[0] > content_width+border:
                        print("-------\nWarning\nContent is overflowing to the right of the page.\n-------")
                    pointer[0] += spacing
                    cw = content_width-(pointer[0]+border)

                # adjust pointer to maximum height for next row to be drawn
                pointer[1] = poih-poihmax

            else:
                pointer = draw_paragraph("Unknown content of {} passed.".format(str(type(row))), pdf,
                               cw, ch, pointer)
            # reset width pointer for next row, raise overflow warning,
            # add spacing and recalculate leftover content height
            pointer[0] = border
            if pointer[1] < border:
                print("-------\nWarning\nContent is overflowing at the bottom of the page.\n-------")
            pointer[1] -= spacing
            ch = pointer[1]-border

    else:
        draw_paragraph("Unknown content of {} passed.".format(str(type(content))), pdf,
                       content_width, content_height, pointer, centered_vertically=True)



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/proteolytic_cleavage.ipynb (unless otherwise specified).

__all__ = ['protease_dict', 'get_cleavage_sites']

# Cell
protease_dict = dict()

protease_dict["arg-c"] = "R"
protease_dict["asp-n"] = "\w(?=D)"
protease_dict["bnps-skatole"] = "W"
protease_dict["caspase 1"] = "(?<=[FWYL]\w[HAT])D(?=[^PEDQKR])"
protease_dict["caspase 2"] = "(?<=DVA)D(?=[^PEDQKR])"
protease_dict["caspase 3"] = "(?<=DMQ)D(?=[^PEDQKR])"
protease_dict["caspase 4"] = "(?<=LEV)D(?=[^PEDQKR])"
protease_dict["caspase 5"] = "(?<=[LW]EH)D"
protease_dict["caspase 6"] = "(?<=VE[HI])D(?=[^PEDQKR])"
protease_dict["caspase 7"] = "(?<=DEV)D(?=[^PEDQKR])"
protease_dict["caspase 8"] = "(?<=[IL]ET)D(?=[^PEDQKR])"
protease_dict["caspase 9"] = "(?<=LEH)D"
protease_dict["caspase 10"] = "(?<=IEA)D"
protease_dict["chymotrypsin high specificity"] = "([FY](?=[^P]))|(W(?=[^MP]))"
protease_dict["chymotrypsin low specificity"] = "([FLY](?=[^P]))|(W(?=[^MP]))|(M(?=[^PY]))|(H(?=[^DMPW]))"
protease_dict["clostripain"] = "R"
protease_dict["cnbr"] = "M"
protease_dict["enterokinase"] = "(?<=[DE]{3})K"
protease_dict["factor xa"] = "(?<=[AFGILTVM][DE]G)R"
protease_dict["formic acid"] = "D"
protease_dict["glutamyl endopeptidase"] = "E"
protease_dict["granzyme b"] = "(?<=IEP)D"
protease_dict["hydroxylamine"] = "N(?=G)"
protease_dict["iodosobenzoic acid"] = "W"
protease_dict["lysc"] = "K"
protease_dict["ntcb"] = "\w(?=C)"
protease_dict["pepsin ph1.3"] = "((?<=[^HKR][^P])[^R](?=[FL][^P]))|((?<=[^HKR][^P])[FL](?=\w[^P]))"
protease_dict["pepsin ph2.0"] = "((?<=[^HKR][^P])[^R](?=[FLWY][^P]))|((?<=[^HKR][^P])[FLWY](?=\w[^P]))"
protease_dict["proline endopeptidase"] = "(?<=[HKR])P(?=[^P])"
protease_dict["proteinase k"] = "[AEFILTVWY]"
protease_dict["staphylococcal peptidase i"] = "(?<=[^E])E"
protease_dict["thermolysin"] = "[^DE](?=[AFILMV])"
protease_dict["thrombin"] = "((?<=G)R(?=G))|((?<=[AFGILTVM][AFGILTVWA]P)R(?=[^DE][^DE]))"
protease_dict["trypsin_full"] = "([KR](?=[^P]))|((?<=W)K(?=P))|((?<=M)R(?=P))"
protease_dict["trypsin_exception"] = "((?<=[CD])K(?=D))|((?<=C)K(?=[HY]))|((?<=C)R(?=K))|((?<=R)R(?=[HR]))"
protease_dict["non-specific"] = "()"
protease_dict["trypsin"] = "([KR](?=[^P]))"
protease_dict["custom_enzyme"] = "[]"

# Cell
import re
def get_cleavage_sites(sequence: str, protease: str):
    """
    Function to get the position of proteolytic cleavage sites in a sequence.

    Args:
        sequence (str): Amino acid sequence.
        protease (str): Protease to use for in silico digestion.
    Returns:
        list: List of cleavage site indices for the selected protease.

    """
    try:
        pattern = re.compile(protease_dict[protease])
    except:
        return []
    pattern_idx = pattern.finditer(sequence)
    pattern_idx = [m.start(0) for m in pattern_idx]
    return pattern_idx


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/Preprocessing.ipynb (unless otherwise specified).

__all__ = ['extract_uniprot_id', 'expand_protein_ids', 'pep_position_helper', 'get_peptide_position', 'get_ptm_sites',
           'get_modifications', 'format_input_data']

# Cell
import pandas as pd
def extract_uniprot_id(protein_id:str):
    """
    Extract the Uniprot unique entry id from the unusual formatted protein_id.
    """
    if 'sp' in protein_id:
        return protein_id.split('|')[1]
    elif '__' in protein_id:
        return protein_id.split('__')[-1]
    return protein_id

def expand_protein_ids(df: pd.DataFrame):
    """
    Function to split protein groups in 'all_protein_ids' by ';' into separate rows.
    The resulting dataframe has a new column 'unique_protein_id'.
    Args:
        df (pd.DataFrame): Experimental data that was imported by the 'import_data' function.
    Returns:
        pd.DataFrame: Exploded dataframe with a new column 'unique_protein_id'.
    """
    df = df.copy(deep=True)
    df.all_protein_ids = df.all_protein_ids.str.split(';')
    df["all_protein_ids_all"] = df.all_protein_ids.apply(lambda x: ';'.join(sorted(x)))
    res = df.explode('all_protein_ids').reset_index(drop=True)
    res.columns = ['unique_protein_id','modified_sequence','naked_sequence','all_protein_ids']
    res.unique_protein_id = res.unique_protein_id.apply(lambda x: extract_uniprot_id(x))
    return res

# Cell
import re
import numpy as np
from pyteomics import fasta

def pep_position_helper(seq: str, prot: str, fasta: fasta, verbose: bool = True):
    """
    Helper function for 'get_peptide_position'.

    Args:
        seq (str): Naked peptide sequence.
        prot (str): UniProt protein accession.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        verbose (bool, optional): Flag to print warnings if no matching sequence is found for a protein in the provided fasta. Defaults to 'True'.
    Returns:
        [int, int]: int: peptide start position, int: peptide end position

    """
    try:
        fasta_prot = fasta[prot]
        search_res = re.search(seq,fasta_prot.sequence)
        if search_res is None:
            start, end = np.NaN, np.NaN
            if verbose:
                warnings.warn(f'Peptide sequence {seq} could not be mached to {prot} in the selected fasta.')
        else:
            start, end = search_res.span()
    except:
        start, end = np.NaN, np.NaN
        if verbose:
            warnings.warn(f'No matching entry for {prot} in the selected fasta.')

    return start, end-1

# Cell

import warnings

def get_peptide_position(df: pd.DataFrame, fasta: fasta, verbose:bool = True):
    """
    Function to get start and end position of each peptide in the given protein.

    Args:
        df (pd.DataFrame): Experimental data that was imported by the 'import_data' function and processed by 'expand_protein_ids'.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        verbose (bool, optional): Flag to print warnings if no matching sequence is found for a protein in the provided fasta. Defaults to 'True'.
    Returns:
        pd.DataFrame: Dataframe with a new columns 'start' and 'end', indicating the start and end index position of the peptide sequence.

    """
    res = df.copy(deep=True)
    res[['start','end']] = res.apply(lambda row: pep_position_helper(row['naked_sequence'],
                                                                     row['unique_protein_id'],
                                                                     fasta,
                                                                     verbose=verbose),
                                     axis=1, result_type='expand')

    res_na = res[res.isnull().any(1)]
    prots_na = res_na.unique_protein_id.unique()

    res = res.dropna()
    res['start'] = res['start'].astype('int64')
    res['end'] = res['end'].astype('int64')
    return res

# Cell
import numpy as np

def get_ptm_sites(peptide: str, modification_reg: str):
    """
    Function to get sequence positions of all PTMs of a peptide in the given protein.

    Args:
        peptide (str): Experimental data that was imported by the 'import_data' function and processed by 'expand_protein_ids'.
        modification_reg (str): Regular expression for the modifications.
    Returns:
        list: List of integers with PTM site location indices on the peptide.

    """
    r = re.compile(modification_reg)
    starts=[]
    ends=[]
    for m in r.finditer(peptide):
        starts.append(m.start())
        ends.append(m.end())
    PTM_sites = np.zeros(len(starts))
    for idx in range(0,len(starts)):
        if idx > 0:
            previous_len=previous_len+(ends[idx-1]-starts[idx-1])
        else:
            previous_len=0
        PTM_sites[idx] = starts[idx] - previous_len - 1
    PTM_sites = [int(i)+1 if i < 0 else int(i) for i in PTM_sites]
    return PTM_sites

# Cell
import re

def get_modifications(df: pd.DataFrame, mod_reg: str):
    """
    Function to get sequence positions and modification types of all PTMs of a peptide in the given protein.

    Args:
        df (pd.DataFrame): Experimental data that was imported by the 'import_data' function and processed by 'expand_protein_ids'.
        mod_reg (str): Regular expression for the modifications.
    Returns:
        pd.DataFrame: Dataframe with a new columns 'PTMsites' and 'PTMtypes' containing lists of PTM site indices and modification types, respectively.

    """
    res = df.copy(deep=True)
    res['PTMsites'] = res.apply(lambda row: get_ptm_sites(row['modified_sequence'],
                                                        modification_reg=mod_reg), axis=1)
    res['PTMtypes'] = res.apply(lambda row: re.findall(mod_reg, row['modified_sequence']), axis=1)
    return res

# Cell

def format_input_data(df: pd.DataFrame, fasta: fasta, modification_exp: str, verbose:bool = True):
    """
    Function to format input data and to annotate sequence start and end positions plus PTM sites.

    Args:
        df (pd.DataFrame): Experimental data that was imported by the 'import_data' function.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        modification_exp (str): Regular expression for the modifications.
        verbose (bool, optional): Flag to print warnings if no matching sequence is found for a protein in the provided fasta. Defaults to 'True'.
    Returns:
        pd.DataFrame: Dataframe with unique uniprot accessions, sequence start and end positions, and PTM site information.

    """
    res = df.copy(deep=True)
    res = expand_protein_ids(res)
    res = get_peptide_position(res, fasta = fasta, verbose=verbose)
    res = get_modifications(res, mod_reg = modification_exp)
    return res